#include <fcntl.h>
#include <pty.h>
#include <unistd.h>
#include <sys/wait.h>
#include <iostream>
#include <cstring>
#include <vector>
#include <sstream>
#include <iterator>
#include <iomanip>
#include <ctime>
#include <chrono>
#include <filesystem>

#include <poll.h>
#include <termios.h>  // Для работы с терминальными атрибутами

enum OutputType {
    STDOUT,
    STDERR
};

struct WaitResult {
    int error_code;
    int index;
    OutputType output_type;
    std::string stdout_data;
    std::string stderr_data;
};

struct FileInfo {
    int permissions;
    std::string owner;
    std::string group;
    long size;
    std::filesystem::file_time_type modified_time;
    std::filesystem::path path;
    std::filesystem::path symlink_path;
    bool is_directory;
};

class FISHClient {
    int master_fd;
    int stderr_pipe[2];
    pid_t pid;

public:

    FISHClient() : master_fd(-1), pid(-1) {}

	bool openApp(const char* app, const char* arg) {
	    if (pipe(stderr_pipe) < 0) {
	        std::cerr << "pipe failed\n";
	        return false;
	    }

	    pid = forkpty(&master_fd, nullptr, nullptr, nullptr);

	    if (pid < 0) {
	        std::cerr << "forkpty failed\n";
	        return false;
	    } else if (pid == 0) {
	        // Child process

			// ***
	        // Получаем текущие атрибуты терминала
	        struct termios term;
	        if (tcgetattr(STDIN_FILENO, &term) == -1) {
	            std::cerr << "tcgetattr failed\n";
	            return false;
	        }

	        // Отключаем эхо
	        term.c_lflag &= ~ECHO;

	        // Применяем новые атрибуты
	        if (tcsetattr(STDIN_FILENO, TCSANOW, &term) == -1) {
	            std::cerr << "tcsetattr failed\n";
	            return false;
	        }
			// ***

	        close(stderr_pipe[0]); // close read end
	        dup2(stderr_pipe[1], STDERR_FILENO); // redirect stderr to the pipe
	        close(stderr_pipe[1]); // close write end, as it's now duplicated
	        execlp(app, app, arg, (char*) nullptr);
	        std::cerr << "execlp failed\n";

	        return false;
	    }

	    // Parent process
	    close(stderr_pipe[1]); // close write end

	    // Now you can read from stderr_pipe[0] to get the stderr output
	    // and from master_fd to get the stdout output

		// Учтём, что в fish_pipeopen, на основе которого написана эта функция, stderr
		// просто отбрасывается.
		// Так что мы, по идее, могли бы использовать и предыдущий (проверенный) вариант openApp(),
		// где stdout и stderr смешиваются:
		/*
		    bool openApp(const char* app, const char* arg) {
		        pid = forkpty(&master_fd, nullptr, nullptr, nullptr);

		        if (pid < 0) {
		            std::cerr << "forkpty failed\n";
		            return false;
		        } else if (pid == 0) {
		            // Child process
		            execlp(app, app, arg, (char*) nullptr);
		            std::cerr << "execlp failed\n";
		            return false;
		        }

		        return true;
		    }
		*/

	    return true;
	}

	WaitResult waitFor(const std::vector<std::string>& commands, const std::vector<std::string>& expectedStrings) {
	    WaitResult result;
	    result.error_code = 0;

	    for (const auto& command : commands) {
	        if (write(master_fd, command.c_str(), command.length()) < 0) {
	            result.error_code = errno;
	            return result;
	        }
	    }

	    struct pollfd fds[2];
	    fds[0].fd = master_fd;
	    fds[0].events = POLLIN;
	    fds[1].fd = stderr_pipe[0];
	    fds[1].events = POLLIN;

	    std::string stdout_buffer, stderr_buffer;

	    while (true) {
	        int ret = poll(fds, 2, -1);

	        if (ret > 0) {
	            char buffer[1024];

	            if (fds[0].revents & POLLIN) {
	                int n = read(master_fd, buffer, sizeof(buffer));
	                if (n > 0) {
	                    stdout_buffer.append(buffer, n);
	                    //std::cout << "Debug (stdout): " << buffer << std::endl;  // ***
	                }
	            }

	            if (fds[1].revents & POLLIN) {
	                int n = read(stderr_pipe[0], buffer, sizeof(buffer));
	                if (n > 0) {
	                    stderr_buffer.append(buffer, n);
	                    //std::cout << "Debug (stderr): " << buffer << std::endl;  // ***
	                }
	            }

	            for (size_t i = 0; i < expectedStrings.size(); ++i) {
	                if (stdout_buffer.find(expectedStrings[i]) != std::string::npos) {
	                    result.index = i;
	                    result.output_type = STDOUT;
	                    result.stdout_data = stdout_buffer;
	                    result.stderr_data = stderr_buffer;
	                    return result;
	                }

	                if (stderr_buffer.find(expectedStrings[i]) != std::string::npos) {
	                    result.index = i;
	                    result.output_type = STDERR;
	                    result.stdout_data = stdout_buffer;
	                    result.stderr_data = stderr_buffer;
	                    return result;
	                }
	            }
	        } else if (ret < 0) {
	            result.error_code = errno;
	            result.stdout_data = stdout_buffer;
	            result.stderr_data = stderr_buffer;
	            return result;
	        }
	    }
	}

	std::vector<FileInfo> parseFishLs(const std::string& buffer) {
	    std::vector<FileInfo> files;
	    std::istringstream stream(buffer);
	    std::string line;
	    FileInfo fileInfo = {0};

	    while (std::getline(stream, line)) {
	        if (line.empty()) continue;

	        char type = line[0];
	        std::string data = line.substr(1);
					
			//printf("*** %s\n", line.c_str());

	        switch (type) {
	            case ':': {

	                if (!fileInfo.path.empty()) {
	                    files.push_back(fileInfo);
	                }
	                fileInfo = FileInfo();
	                
	                // Парсинг имени файла и пути символической ссылки
	                size_t arrow_pos = data.find(" -> ");
	                if (arrow_pos != std::string::npos) {
	                    fileInfo.path = data.substr(0, arrow_pos);
	                    fileInfo.symlink_path = data.substr(arrow_pos + 4);
	                } else {
	                    fileInfo.path = data;
	                }

	                break;
				}

	            case 'S':
	                fileInfo.size = std::stol(data);
	                break;

	            // ... другие case ...

	            default:
	                break;
	        }
	    }

	    if (!fileInfo.path.empty()) {
			//printf("***2 %s\n", fileInfo.path.c_str());
	        files.push_back(fileInfo);
	    }

	    return files;
	}

    ~FISHClient() {
       if (pid > 0) {
            // Проверяем, завершен ли дочерний процесс
            int status;
            if (waitpid(pid, &status, WNOHANG) == 0) {
                // Если нет, отправляем сигнал для завершения
                kill(pid, SIGTERM);
                waitpid(pid, &status, 0); // Ожидаем завершения
            }
            std::cout << "Child exited with status " << status << '\n';
        }

	    // Закрыть файловые дескрипторы
	    if (master_fd != -1) {
	        close(master_fd);
	    }
	    if (stderr_pipe[0] != -1) {
	        close(stderr_pipe[0]);
	    }
    }
};

